import { Role } from './ad.types';
import { Target } from './candidate.types';
import { CarpoolPathCreatorException } from './match.errors';
import { Actor } from './value-objects/actor.value-object';
import { Point } from './value-objects/point.value-object';
import { CarpoolPathItem } from './value-objects/carpool-path-item.value-object';

export class CarpoolPathCreator {
  private PRECISION = 5;

  constructor(
    private readonly driverWaypoints: Point[],
    private readonly passengerWaypoints: Point[],
  ) {
    if (driverWaypoints.length < 2)
      throw new CarpoolPathCreatorException(
        new Error('At least 2 driver waypoints must be defined'),
      );
    if (passengerWaypoints.length < 2)
      throw new CarpoolPathCreatorException(
        new Error('At least 2 passenger waypoints must be defined'),
      );
  }

  /**
   * Creates a path (a list of carpoolPathItem) between driver waypoints
      and passenger waypoints respecting the order
      of the driver waypoints
      Inspired by :
      https://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment
   */
  public carpoolPath = (): CarpoolPathItem[] =>
    this._consolidate(
      this._mixedCarpoolPath(
        this._driverCarpoolPath(),
        this._passengerCarpoolPath(),
      ),
    );

  private _mixedCarpoolPath = (
    driverCarpoolPath: CarpoolPathItem[],
    passengerCarpoolPath: CarpoolPathItem[],
  ): CarpoolPathItem[] =>
    driverCarpoolPath.length == 2
      ? this._simpleMixedCarpoolPath(driverCarpoolPath, passengerCarpoolPath)
      : this._complexMixedCarpoolPath(driverCarpoolPath, passengerCarpoolPath);

  private _driverCarpoolPath = (): CarpoolPathItem[] =>
    this.driverWaypoints.map(
      (waypoint: Point, index: number) =>
        new CarpoolPathItem({
          lon: waypoint.lon,
          lat: waypoint.lat,
          actors: [
            new Actor({
              role: Role.DRIVER,
              target: this._getTarget(index, this.driverWaypoints),
            }),
          ],
        }),
    );

  /**
   * Creates the passenger carpoolPath with original passenger waypoints, adding driver waypoints that are the same
   */
  private _passengerCarpoolPath = (): CarpoolPathItem[] => {
    const carpoolPath: CarpoolPathItem[] = [];
    this.passengerWaypoints.forEach(
      (passengerWaypoint: Point, index: number) => {
        const carpoolPathItem: CarpoolPathItem = new CarpoolPathItem({
          lon: passengerWaypoint.lon,
          lat: passengerWaypoint.lat,
          actors: [
            new Actor({
              role: Role.PASSENGER,
              target: this._getTarget(index, this.passengerWaypoints),
            }),
          ],
        });
        if (
          this.driverWaypoints.filter((driverWaypoint: Point) =>
            passengerWaypoint.equals(driverWaypoint),
          ).length == 0
        ) {
          carpoolPathItem.actors.push(
            new Actor({
              role: Role.DRIVER,
              target: Target.NEUTRAL,
            }),
          );
        }
        carpoolPath.push(carpoolPathItem);
      },
    );
    return carpoolPath;
  };

  private _simpleMixedCarpoolPath = (
    driverCarpoolPath: CarpoolPathItem[],
    passengerCarpoolPath: CarpoolPathItem[],
  ): CarpoolPathItem[] => [
    driverCarpoolPath[0],
    ...passengerCarpoolPath,
    driverCarpoolPath[1],
  ];

  private _complexMixedCarpoolPath = (
    driverCarpoolPath: CarpoolPathItem[],
    passengerCarpoolPath: CarpoolPathItem[],
  ): CarpoolPathItem[] => {
    let mixedCarpoolPath: CarpoolPathItem[] = [...driverCarpoolPath];
    const originInsertIndex: number = this._insertIndex(
      passengerCarpoolPath[0],
      driverCarpoolPath,
    );
    mixedCarpoolPath = [
      ...mixedCarpoolPath.slice(0, originInsertIndex),
      passengerCarpoolPath[0],
      ...mixedCarpoolPath.slice(originInsertIndex),
    ];
    const destinationInsertIndex: number =
      this._insertIndex(
        passengerCarpoolPath[passengerCarpoolPath.length - 1],
        driverCarpoolPath,
      ) + 1;
    mixedCarpoolPath = [
      ...mixedCarpoolPath.slice(0, destinationInsertIndex),
      passengerCarpoolPath[passengerCarpoolPath.length - 1],
      ...mixedCarpoolPath.slice(destinationInsertIndex),
    ];
    return mixedCarpoolPath;
  };

  private _insertIndex = (
    targetCarpoolPathItem: CarpoolPathItem,
    carpoolPath: CarpoolPathItem[],
  ): number =>
    this._closestSegmentIndex(
      targetCarpoolPathItem,
      this._segments(carpoolPath),
    ) + 1;

  private _segments = (carpoolPath: CarpoolPathItem[]): CarpoolPathItem[][] => {
    const segments: CarpoolPathItem[][] = [];
    carpoolPath.forEach((carpoolPathItem: CarpoolPathItem, index: number) => {
      if (index < carpoolPath.length - 1)
        segments.push([carpoolPathItem, carpoolPath[index + 1]]);
    });
    return segments;
  };

  private _closestSegmentIndex = (
    carpoolPathItem: CarpoolPathItem,
    segments: CarpoolPathItem[][],
  ): number => {
    const distances: Map<number, number> = new Map();
    segments.forEach((segment: CarpoolPathItem[], index: number) => {
      distances.set(index, this._distanceToSegment(carpoolPathItem, segment));
    });
    const sortedDistances: Map<number, number> = new Map(
      [...distances.entries()].sort((a, b) => a[1] - b[1]),
    );
    const [closestSegmentIndex] = sortedDistances.keys();
    return closestSegmentIndex;
  };

  private _distanceToSegment = (
    carpoolPathItem: CarpoolPathItem,
    segment: CarpoolPathItem[],
  ): number =>
    parseFloat(
      Math.sqrt(
        this._distanceToSegmentSquared(carpoolPathItem, segment),
      ).toFixed(this.PRECISION),
    );

  private _distanceToSegmentSquared = (
    carpoolPathItem: CarpoolPathItem,
    segment: CarpoolPathItem[],
  ): number => {
    const length2: number = this._distanceSquared(
      new Point({
        lon: segment[0].lon,
        lat: segment[0].lat,
      }),
      new Point({
        lon: segment[1].lon,
        lat: segment[1].lat,
      }),
    );
    if (length2 == 0)
      return this._distanceSquared(
        new Point({
          lon: carpoolPathItem.lon,
          lat: carpoolPathItem.lat,
        }),
        new Point({
          lon: segment[0].lon,
          lat: segment[0].lat,
        }),
      );
    const length: number = Math.max(
      0,
      Math.min(
        1,
        ((carpoolPathItem.lon - segment[0].lon) *
          (segment[1].lon - segment[0].lon) +
          (carpoolPathItem.lat - segment[0].lat) *
            (segment[1].lat - segment[0].lat)) /
          length2,
      ),
    );
    const newPoint: Point = new Point({
      lon: segment[0].lon + length * (segment[1].lon - segment[0].lon),
      lat: segment[0].lat + length * (segment[1].lat - segment[0].lat),
    });
    return this._distanceSquared(
      new Point({
        lon: carpoolPathItem.lon,
        lat: carpoolPathItem.lat,
      }),
      newPoint,
    );
  };

  private _distanceSquared = (point1: Point, point2: Point): number =>
    parseFloat(
      (
        Math.pow(point1.lon - point2.lon, 2) +
        Math.pow(point1.lat - point2.lat, 2)
      ).toFixed(this.PRECISION),
    );

  private _getTarget = (index: number, waypoints: Point[]): Target =>
    index == 0
      ? Target.START
      : index == waypoints.length - 1
      ? Target.FINISH
      : Target.INTERMEDIATE;

  /**
   * Consolidate carpoolPath by removing duplicate actors (eg. driver with neutral and start or finish target)
   */
  private _consolidate = (
    carpoolPath: CarpoolPathItem[],
  ): CarpoolPathItem[] => {
    const uniquePoints: Point[] = [];
    carpoolPath.forEach((carpoolPathItem: CarpoolPathItem) => {
      if (
        uniquePoints.find((point: Point) =>
          point.equals(
            new Point({
              lon: carpoolPathItem.lon,
              lat: carpoolPathItem.lat,
            }),
          ),
        ) === undefined
      )
        uniquePoints.push(
          new Point({
            lon: carpoolPathItem.lon,
            lat: carpoolPathItem.lat,
          }),
        );
    });
    return uniquePoints.map(
      (point: Point) =>
        new CarpoolPathItem({
          lon: point.lon,
          lat: point.lat,
          actors: carpoolPath
            .filter((carpoolPathItem: CarpoolPathItem) =>
              new Point({
                lon: carpoolPathItem.lon,
                lat: carpoolPathItem.lat,
              }).equals(point),
            )
            .map((carpoolPathItem: CarpoolPathItem) => carpoolPathItem.actors)
            .flat(),
        }),
    );
  };
}
